X(FACADE_SLICE)
X(A0)
X(B0)
X(C0)
X(D0)
X(A1)
X(B1)
X(C1)
X(D1)
X(M0)
X(M1)
X(FCI)
X(FXA)
X(FXB)
X(CLK)
X(LSR)
X(CE)
X(DI0)
X(DI1)
X(WD0)
X(WD1)
X(WAD0)
X(WAD1)
X(WAD2)
X(WAD3)
X(WRE)
X(WCK)
X(F0)
X(Q0)
X(F1)
X(Q1)
X(FCO)
X(OFX0)
X(OFX1)
X(WDO0)
X(WDO1)
X(WDO2)
X(WDO3)
X(WADO0)
X(WADO1)
X(WADO2)
X(WADO3)

X(MODE)
X(GSR)
X(SRMODE)
X(CEMUX)
X(CLKMUX)
X(LSRMUX)
X(LSRONMUX)
X(LUT0_INITVAL)
X(LUT1_INITVAL)
X(REGMODE)
X(REG0_SD)
X(REG1_SD)
X(REG0_REGSET)
X(REG1_REGSET)
X(CCU2_INJECT1_0)
X(CCU2_INJECT1_1)
X(WREMUX)


X(FACADE_FF)
X(DI)
X(Q)

X(REGSET)


X(FACADE_IO)
X(PAD)
X(I)
X(EN)
X(O)

X(DIR)


X(LUT4)
X(A)
X(B)
X(C)
X(D)
X(Z)

X(INIT)


X(PFUMX)
X(ALUT)
X(BLUT)


X(L6MUX21)
X(SD)


X(T)
X(IOLDO)
X(IOLTO)


X(OSCH)
X(STDBY)
X(OSC)
X(SEDSTDBY)


X(DCCA)
X(CLKI)
X(CLKO)


X(DCMA)
X(CLK0)
X(CLK1)
X(SEL)
X(DCMOUT)

X(BEL)
X(GND)
X(IO_TYPE)
X(LOC)
X(NOM_FREQ)
X(VCC)
X(lcmxo2_1200hc)
X(lcmxo2_2000hc)
X(lcmxo2_256hc)
X(lcmxo2_4000hc)
X(lcmxo2_640hc)
X(lcmxo2_7000hc)
X(machxo2)
X(pack)
X(place)
X(placer)
X(route)
X(router)
